import abc
import itertools
import jinja2

from raco import algebra
from raco import expression
from raco import catalog
from raco.algebra import gensym
from raco.expression import UnnamedAttributeRef
from raco.backends import Language
from raco.pipelines import Pipelined
from raco import types

import logging
from functools import reduce
_LOG = logging.getLogger(__name__)


_PACKAGE_PATH = 'raco.backends'


def prepend_loader(env, loader):
    newenv = env.overlay(loader=jinja2.ChoiceLoader([loader, env.loader]))
    # newenv = jinja2.Environment(
    #    loader=jinja2.ChoiceLoader([loader, env.loader]))
    return newenv


def prepend_template_relpath(env, relpath):
    return prepend_loader(env, jinja2.PackageLoader(_PACKAGE_PATH, relpath))


class CBaseLanguage(Language):
    _external_indexing = False

    @classmethod
    def set_external_indexing(cls, b):
        cls._external_indexing = b

    @classmethod
    def c_stringify(cls, st):
        """ turn " in the string into \" since C ' are chars
        """
        return '"{0}"'.format(st.value.replace('"', '\\"'))

    @staticmethod
    def __get_env_for_template_libraries__(*libraries):
        """
        create a Jinja2 Environment with loaders for provided
         libraries and base C templates

        @param libraries: env will load templates from these libraries in order
        @return: a jinja2.Environment
        """
        child_loaders = [
            jinja2.PackageLoader(_PACKAGE_PATH, l) for l in libraries]
        loaders = child_loaders + \
            [jinja2.PackageLoader(_PACKAGE_PATH, 'cpp/cbase_templates')]

        # StrictUndefined makes uses of the result of render() fail when
        # a template variable is undefined, which is most useful for debugging
        return jinja2.Environment(
            undefined=jinja2.StrictUndefined,
            loader=jinja2.ChoiceLoader(loaders))

    @classmethod
    @abc.abstractmethod
    def cgenv(cls):
        pass

    @classmethod
    def body(cls, compileResult):
        queryexec = compileResult.getExecutionCode()
        initialized = compileResult.getInitCode()
        declarations = compileResult.getDeclCode()
        cleanups = compileResult.getCleanupCode()
        resultsym = "__result__"
        return cls.base_template().render(locals())

    @classmethod
    @abc.abstractmethod
    def base_template(cls):
        pass

    @staticmethod
    def comment(txt):
        return "// %s\n" % txt

    nextstrid = 0

    @classmethod
    def newstringident(cls):
        r = """str_%s""" % (cls.nextstrid)
        cls.nextstrid += 1
        return r

    @classmethod
    def compile_numericliteral(cls, value):
        return '%s' % (value), [], []

    @classmethod
    def negation(cls, input):
        innerexpr, decls, inits = input
        return "(!%s)" % (innerexpr,), decls, inits

    @classmethod
    def negative(cls, input):
        innerexpr, decls, inits = input
        return "(-%s)" % (innerexpr,), decls, inits

    @staticmethod
    def _extract_code_decl_init(args):
        if len(args) == 0:
            return [], [], []
        else:
            codes = [c for c, _, _ in args]
            decls_combined = reduce(lambda sofar, x: sofar + x,
                                    [d for _, d, _ in args])
            inits_combined = reduce(lambda sofar, x: sofar + x,
                                    [i for _, _, i in args])
            return codes, decls_combined, inits_combined

    @classmethod
    def expression_combine(cls, args, operator="&&"):
        codes, decls, inits = cls._extract_code_decl_init(args)

        # special case for integer divide. C doesn't have this syntax
        # Rely on automatic conversion from float to int
        this_decls = []
        this_inits = []
        if operator == "//":
            operator = "/"
        # special case for string LIKE, use overloaded mod operator
        elif operator == "like":
            # NOTE: LIKE probably shouldn't be implemented as
            # a "binop" because the input type != output type
            assert len(args) == 2, "LIKE only combines 2 arguments"
            operator = "%"

            # hoist pattern compilation out of the loop processing
            # Unchecked precondition: codes[1] is independent of the tuple
            varname = gensym()
            this_decls.append("std::regex {var};\n".format(var=varname))
            this_inits.append(cls.on_all(
                """{var} = compile_like_pattern({str});
                """.format(var=varname, str=codes[1])))
            # replace the string literal with the regex
            codes[1] = varname

        opstr = " %s " % operator
        conjunc = opstr.join(["(%s)" % c for c in codes])
        _LOG.debug("conjunc: %s", conjunc)
        return "( %s )" % conjunc, \
               decls + this_decls, \
               inits + this_inits

    @classmethod
    def on_all(cls, code):
        """Parallel on all partitions"""
        return code

    @classmethod
    def function_call(cls, name, *args, **kwargs):
        is_custom = kwargs.get('custom', False)
        if not is_custom:
            name = name.lower()

        codes, decls, inits = cls._extract_code_decl_init(list(args))
        argscode = ",".join(["{0}".format(d) for d in codes])

        # special cases where name is not just the name,
        # for example there is a namespace preceding it
        name = {'year': 'dates::year',
                'abs': 'std::abs'}.get(name, name)

        code = "{name}({argscode})".format(
            name=name, argscode=argscode)
        return code, decls, inits

    @classmethod
    def typename(cls, raco_type, allow_subs=True):
        # if external indexing is on, make strings into ints
        if cls._external_indexing and \
                raco_type == types.STRING_TYPE and \
                allow_subs:
            raco_type = types.LONG_TYPE

        n = {
            types.LONG_TYPE: 'int64_t',
            types.BOOLEAN_TYPE: 'bool',
            types.DOUBLE_TYPE: 'double',
            types.STRING_TYPE: 'std::array<char, MAX_STR_LEN>'
        }.get(raco_type)

        assert n is not None, \
            "Clang does not yet support type {type}".format(type=n)
        return n

    @classmethod
    def cast(cls, castto, inputexpr):
        inputcode, decls, inits = inputexpr
        typen = cls.typename(castto)
        code = "(static_cast<{typename}>({expr}))".format(
            typename=typen, expr=inputcode)
        return code, decls, inits

    @classmethod
    def compile_attribute(cls, expr, **kwargs):
        if isinstance(expr, expression.NamedAttributeRef):
            raise TypeError(
                "Error compiling attribute reference %s. \
                C compiler only support unnamed perspective. \
                Use helper function unnamed." % expr)
        if isinstance(expr, expression.UnnamedAttributeRef):
            tupleref = kwargs.get('tupleref')
            assert tupleref is not None, \
                "Cannot compile {0} without a tupleref".format(expr)

            position = expr.position
            assert position >= 0
            return tupleref.get_code(position), [], []
        if isinstance(expr, expression.NamedStateAttributeRef):
            state_scheme = kwargs.get('state_scheme')
            assert state_scheme is not None, \
                "Cannot compile {0} without a state_scheme".format(expr)

            # A bit hacky but simple: when we
            # haven't physically joined left and right in sides in
            # binary operators into one state scheme,
            # we can treat them separately.
            # Note that it assumes the full scheme is just
            # state_scheme + state_scheme
            if hasattr(expr, 'tagged_state_id'):
                name_id = expr.tagged_state_id
            else:
                name_id = ''

            position = expr.get_position(None, state_scheme)
            code = StagedTupleRef.get_code_with_name(
                position, "state{0}".format(name_id))
            return code, [], []

        assert False, "{expr} is unsupported attribute".format(expr=expr)

    @classmethod
    def compile_stringliteral(cls, s):
        return '(%s)' % s, [], []

    @classmethod
    def ifelse(cls, when_compiled, else_compiled):
        if_template = """
        if ({cond}) {{
        {then}
        }}
        """

        else_template = """
        else {{
        {then}
        }}
        """

        return cls._conditional_(when_compiled, else_compiled,
                                 if_template, else_template, "else")

    @classmethod
    def limits(cls, side, typ):
        return "std::numeric_limits<{type}>::{side}()".format(
            type=cls.typename(typ),
            side=side)

    @classmethod
    def conditional(cls, when_compiled, else_compiled):
        if_template = """{cond} ? {then}"""

        else_template = """: {then}"""

        return cls._conditional_(when_compiled, else_compiled,
                                 if_template, else_template, ":")

    @classmethod
    def _conditional_(cls, when_compiled, else_compiled,
                      if_template, else_template, else_joiner):
        def by_pairs(l):
            assert len(l) % 2 == 0, "must be even length"
            for i in range(0, len(l), 2):
                yield l[i], l[i + 1]

        flatten_when_then = list(itertools.chain.from_iterable(when_compiled))
        when_exec, when_decls, when_inits = zip(*flatten_when_then)

        code = else_joiner.join([if_template.format(cond=cond, then=then)
                                 for (cond, then) in by_pairs(when_exec)])

        if else_compiled is not None:
            code += else_template.format(then=else_compiled[0])

        return code, \
            list(itertools.chain.from_iterable(when_decls)) + else_compiled[1], \
            list(itertools.chain.from_iterable(when_inits)) + else_compiled[2]


_cgenv = CBaseLanguage.__get_env_for_template_libraries__()

# TODO:
# The following is actually a staged materialized tuple ref.
# we should also add a staged reference tuple ref that
# just has relationsymbol and row


class StagedTupleRef(object):
    nextid = 0

    @staticmethod
    def get_append(out_tuple_type, type1, type1numfields,
                   type2, type2numfields):

        append_func_name = "create_" + gensym()

        result_type = out_tuple_type
        combine_function_def = _cgenv.get_template(
            "materialized_tuple_create_two.cpp").render(locals())
        return append_func_name, combine_function_def

    @classmethod
    def genname(cls):
        # use StagedTupleRef so everyone shares one mutable copy of nextid
        x = StagedTupleRef.nextid
        StagedTupleRef.nextid += 1
        return "t_%03d" % x

    def __init__(self, relsym, scheme):
        self.name = self.genname()
        self.relsym = relsym
        self.scheme = scheme
        self.__typename = None

    def __str__(self):
        return "Tuple{{name={name}, relsym={relsym}, scheme={scheme}}}".format(
            name=self.name,
            relsym=self.relsym,
            scheme=self.scheme)

    def copy_type(self):
        """
        Create a new tuple ref of the same type but new symbol name
        """
        return self.__class__(self.relsym, self.scheme)

    def getTupleTypename(self):
        if self.__typename is None:
            fields = ""
            relsym = self.relsym
            if len(self.scheme) == 0:
                fields = "ZERO"
            else:
                for i in range(0, len(self.scheme)):
                    fieldnum = i
                    fields += "_%(fieldnum)s" % locals()

            self.__typename = "MaterializedTupleRef_%(relsym)s%(fields)s" \
                              % locals()

        return self.__typename

    @staticmethod
    def get_code_with_name(position, name):
        return "{name}.f{position}".format(position=position, name=name)

    def get_code(self, position):
        return StagedTupleRef.get_code_with_name(position, self.name)

    def set_func_code(self, position):
        return "{name}.f{position}".format(
            position=position, name=self.name)

    def generateDefinition(self):
        template = _cgenv.get_template('materialized_tuple_ref.cpp')

        numfields = len(self.scheme)

        fieldtypes = [CBaseLanguage.typename(t)
                      for t in self.scheme.get_types()]

        string_type_name = CBaseLanguage.typename(types.STRING_TYPE,
                                                  allow_subs=False)

        # stream_sets = emitlist(
        # ["_ret.set<{i}>(std::get<{i}>(_t));".format(i=i)
        #                        for i in range(numfields)])

        additional_code = self.__additionalDefinitionCode__(
            numfields,
            fieldtypes)
        after_def_code = self.__afterDefinitionCode__(numfields, fieldtypes)

        tupletypename = self.getTupleTypename()
        relsym = self.relsym

        code = template.render(locals())
        return code

    def __additionalDefinitionCode__(self, numfields, fieldtypes):
        return ""

    def __afterDefinitionCode__(self, numfields, fieldtypes):
        return ""


class CBaseSelect(Pipelined, algebra.Select):

    def _compile_condition(self, t, state):
        condition_as_unnamed = expression.ensure_unnamed(self.condition, self)

        # compile the predicate into code
        conditioncode, cond_decls, cond_inits = \
            self.language().compile_expression(
                condition_as_unnamed, tupleref=t)
        state.addInitializers(cond_inits)
        state.addDeclarations(cond_decls)
        return conditioncode

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        basic_select_template = _cgenv.get_template('select.cpp')

        conditioncode = self._compile_condition(t, state)

        inner_code_compiled = self.parent().consume(t, self, state)

        code = basic_select_template.render(locals())
        return code


def createTupleTypeConversion(lang, state, input_tuple, result_tuple):
    # add declaration for function to convert from one type to the other
    type1 = input_tuple.getTupleTypename()
    type1numfields = len(input_tuple.scheme)
    convert_func_name = "create_" + gensym()
    result_type = result_tuple.getTupleTypename()
    result_name = result_tuple.name
    input_tuple_name = input_tuple.name
    convert_func = lang._cgenv.get_template(
        'materialized_tuple_create_one.cpp').render(locals())
    state.addDeclarations([convert_func])

    return lang._cgenv.get_template('tuple_type_convert.cpp').render(
        result_type=result_type,
        result_name=result_name,
        convert_func_name=convert_func_name,
        input_tuple_name=input_tuple_name
    )


class CBaseUnionAll(Pipelined, algebra.UnionAll):

    def produce(self, state):
        self.unifiedTupleType = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.unifiedTupleType.generateDefinition()])

        for arg in self.args:
            arg.produce(state)

    def consume(self, t, src, state):
        unified_tuple = self.unifiedTupleType

        assignment_code = \
            createTupleTypeConversion(self.language(),
                                      state,
                                      t,
                                      unified_tuple)

        inner_plan_compiled = \
            self.parent().consume(self.unifiedTupleType, self, state)
        return assignment_code + inner_plan_compiled


class CBaseApply(Pipelined, algebra.Apply):

    def produce(self, state):
        # declare a single new type for project
        # TODO: instead do mark used-columns?

        # always does an assignment to new tuple
        self.newtuple = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.newtuple.generateDefinition()])

        self.input.produce(state)

    def _apply_statements(self, t, state):
        assignment_template = _cgenv.get_template('assignment.cpp')
        dst_name = self.newtuple.name
        dst_type_name = self.newtuple.getTupleTypename()

        code = ""
        for dst_fieldnum, src_label_expr in enumerate(self.emitters):
            dst_set_func = self.newtuple.set_func_code(dst_fieldnum)
            src_label, src_expr = src_label_expr

            # make sure to resolve attribute positions using input schema
            src_expr_unnamed = expression.ensure_unnamed(src_expr, self.input)

            src_expr_compiled, expr_decls, expr_inits = \
                self.language().compile_expression(
                    src_expr_unnamed, tupleref=t)
            state.addInitializers(expr_inits)
            state.addDeclarations(expr_decls)

            code += assignment_template.render(locals())
        return code

    def consume(self, t, src, state):
        code = self.language().comment(self.shortStr())

        dst_name = self.newtuple.name
        dst_type_name = self.newtuple.getTupleTypename()

        # declaration of tuple instance
        code += _cgenv.get_template('tuple_declaration.cpp').render(locals())

        code += self._apply_statements(t, state)

        innercode = self.parent().consume(self.newtuple, self, state)
        code += innercode

        return code


class CBaseProject(Pipelined, algebra.Project):

    def produce(self, state):
        # declare a single new type for project
        # TODO: instead do mark used-columns?

        # always does an assignment to new tuple
        self.newtuple = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.newtuple.generateDefinition()])

        self.input.produce(state)

    def consume(self, t, src, state):
        code = ""

        assignment_template = _cgenv.get_template('assignment.cpp')

        dst_name = self.newtuple.name
        dst_type_name = self.newtuple.getTupleTypename()

        # declaration of tuple instance
        code += _cgenv.get_template('tuple_declaration.cpp').render(locals())

        for dst_fieldnum, src_expr in enumerate(self.columnlist):
            if isinstance(src_expr, UnnamedAttributeRef):
                src_expr_compiled = t.get_code(src_expr.position)
                dst_set_func = t.set_func_code(dst_fieldnum)
            else:
                assert False, "Unsupported Project expression"
            code += assignment_template.render(locals())

        innercode = self.parent().consume(self.newtuple, self, state)
        code += innercode

        return code


from raco.algebra import ZeroaryOperator


class CBaseFileScan(Pipelined, algebra.Scan):

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __get_ascii_scan_template__(self):
        return

    @abc.abstractmethod
    def __get_binary_scan_template__(self):
        return

    def _get_input_aux_decls_template(self):
        return None

    def __get_relation_decl_template__(self, name):
        """
        Implement if the CFileScan implementation requires
        the relation instance to be a global declaration.
        If not then just put the local declaration within
        the *_scan_template.
        """
        return None

    def produce(self, state):

        # Common subexpression elimination
        # don't scan the same file twice
        resultsym = state.lookupExpr(self)
        _LOG.debug("lookup %s(h=%s) => %s", self, self.__hash__(),
                   resultsym)
        if not resultsym:
            # TODO for now this will break
            # whatever relies on self.bound like reusescans
            # Scan is the only place where a relation is declared
            resultsym = gensym()

            name = str(self.relation_key).split(':')[2]
            fstemplate, fsbindings = self.__compileme__(resultsym, name)
            state.saveExpr(self, resultsym)

            stagedTuple = self.new_tuple_ref_for_filescan(
                resultsym,
                self.scheme())
            state.saveTupleDef(resultsym, stagedTuple)

            tuple_type_def = stagedTuple.generateDefinition()
            tuple_type = stagedTuple.getTupleTypename()
            state.addDeclarations([tuple_type_def])

            colnames = self.scheme().get_names()

            rel_decl_template = self.__get_relation_decl_template__(name)
            if rel_decl_template:
                state.addDeclarations([rel_decl_template.render(locals())])

            rel_aux_decl_template = self._get_input_aux_decls_template()

            if rel_aux_decl_template:
                state.addDeclarations([rel_aux_decl_template.render(locals())])

            # now that we have the type, format this in;
            state.setPipelineProperty('type', 'scan')
            state.setPipelineProperty('source', self.__class__)
            state.addPipeline(
                fstemplate.render(fsbindings, result_type=tuple_type))

        # no return value used because parent is a new pipeline
        self.parent().consume(resultsym, self, state)

    def new_tuple_ref_for_filescan(self, resultsym, scheme):
        """instance version of new_tuple_ref.
        Default just calls the cls version"""
        return self.new_tuple_ref(resultsym, scheme)

    def consume(self, t, src, state):
        assert False, "as a source, no need for consume"

    def __compileme__(self, resultsym, name):
        # TODO use the identifiers (don't split str and extract)
        # name = self.relation_key

        _LOG.debug('compiling file scan for relation_key %s'
                   % self.relation_key)

        # tup = (resultsym, self.originalterm.originalorder, self.originalterm)
        # self.trace("// Original query position of %s: term %s (%s)" % tup)

        if isinstance(self.relation_key, catalog.ASCIIFile):
            template = self.__get_ascii_scan_template__()
        else:
            template = self.__get_binary_scan_template__()
        return template, locals()

    def __str__(self):
        return "%s(%s)" % (self.opname(), self.relation_key)

    def __eq__(self, other):
        """
        For what we are using FileScan for, the only use
        of __eq__ is in hashtable lookups for CSE optimization.
        We omit self.schema because the relation_key determines
        the level of equality needed.

        This could break other things, so better may be to
        make a normalized copy of an expression. This could
        include simplification but in the case of Scans make
        the scheme more generic.

        @see MemoryScan.__eq__
        """
        return ZeroaryOperator.__eq__(self, other) and \
            self.relation_key == other.relation_key


# Rules
from raco import rules


class BreakHashJoinConjunction(rules.Rule):

    """A rewrite rule for turning HashJoin(a=c and b=d)
    into select(b=d)[HashJoin(a=c)]"""

    def __init__(self, select_clazz, join_clazz):
        self.select_clazz = select_clazz
        self.join_clazz = join_clazz
        super(BreakHashJoinConjunction, self).__init__()

    def fire(self, expr):
        if isinstance(expr, self.join_clazz) \
                and isinstance(expr.condition.left, expression.EQ) \
                and isinstance(expr.condition.right, expression.EQ):
            return self.select_clazz(expr.condition.right,
                                     self.join_clazz(expr.condition.left,
                                                     expr.left,
                                                     expr.right))

        return expr

    def __str__(self):
        return "%s(a=c and b=d) => %s(b=d)[%s(a=c)]" \
               % (self.join_clazz.__name__,
                  self.select_clazz.__name__,
                  self.join_clazz.__name__)


clang_push_select = [
    rules.SplitSelects(),
    rules.PushSelects(),
    rules.MergeSelects()
]

EMIT_CONSOLE = 'console'
EMIT_FILE = 'file'


class CBaseSink(Pipelined, algebra.Sink):

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        # declare an unused result vector
        resdecl = "std::vector<%s> result;\n" % (t.getTupleTypename())
        state.addDeclarations([resdecl])

        code = self.language().log_unquoted("%s" % t.name, 2)
        return code


class CBaseStore(Pipelined, algebra.Store):

    def __init__(self, emit_print, relation_key, plan):
        super(CBaseStore, self).__init__(relation_key, plan)
        self.emit_print = emit_print

    def produce(self, state):
        self.input.produce(state)

    def _add_result_declaration(self, t, state):
        resdecl = "std::vector<%s> result;\n" % (t.getTupleTypename())
        state.addDeclarations([resdecl])

    def consume(self, t, src, state):
        code = ""
        self._add_result_declaration(t, state)
        code += "result.push_back(%s);\n" % (t.name)

        if self.emit_print == EMIT_CONSOLE:
            code += self.language().log_unquoted("%s" % t.name, 2)
        elif self.emit_print == EMIT_FILE:
            code += self.__file_code__(t, state)

        return code

    @abc.abstractmethod
    def __file_code__(self, t, state):
        pass

    def __repr__(self):
        return "{op}({ep!r}, {rk!r}, {pl!r})".format(op=self.opname(),
                                                     ep=self.emit_print,
                                                     rk=self.relation_key,
                                                     pl=self.input)


class StoreToBaseCStore(rules.Rule):

    """A rule to store tuples into emit_print"""

    def __init__(self, emit_print, subclass):
        self.emit_print = emit_print
        assert issubclass(subclass, CBaseStore), \
            "%s is not a subclass of %s" % (subclass, CBaseStore)
        self.subclass = subclass
        super(StoreToBaseCStore, self).__init__()

    def fire(self, expr):
        if isinstance(expr, algebra.Store):
            return self.subclass(self.emit_print,
                                 expr.relation_key,
                                 expr.input)
        return expr

    def __str__(self):
        return "Store => %s" % self.subclass.__name__


class BaseCGroupby(Pipelined, algebra.GroupBy):

    def __get_initial_value__(self, index, cached_inp_sch=None):
        if cached_inp_sch is None:
            cached_inp_sch = self.input.scheme()

        op = self.aggregate_list[index].__class__.__name__

        # min, max need special values; default to 0 as initial value
        initial_value = {
            'MAX': self.language().limits('min',
                                          self.aggregate_list[index].typeof(
                                              cached_inp_sch, None)),
            'MIN': self.language().limits('max',
                                          self.aggregate_list[index].typeof(
                                              cached_inp_sch, None))
        }.get(op, 0)

        return initial_value




import abc
from raco.utility import emitlist
from algebra import gensym, Operator
import re

import logging
LOG = logging.getLogger(__name__)


class ResolvingSymbol:
    _unique_tag = "_@@UNRESOLVED_SYMBOL_{name}@@_"

    def __init__(self, name):
        self._name = name
        self._placeholder = ResolvingSymbol._mksymbol(name)

    @classmethod
    def _mksymbol(cls, name):
        return cls._unique_tag.format(name=name)

    def getPlaceholder(self):
        return self._placeholder

    def getName(self):
        return self._name

    @classmethod
    def substitute(cls, code, symbols):
        # inefficient multi-string replacement
        # TODO: replace with multi-pattern sed script
        for name, val in symbols.items():
            assert val is not None, "Unresolved symbol: {}".format(name)
            pat = cls._unique_tag.format(name=name)
            code = re.sub(pat, val, code)
        return code


class CompileState(object):

    def __init__(self, lang, cse=True):
        self.language = lang

        self.declarations = []
        self.declarations_later = []
        self.pipelines = []
        self.scan_pipelines = []
        self.flush_pipelines = []
        self.initializers = []
        self.cleanups = []
        self.pipeline_count = 0

        # { expression => symbol for materialized result }
        self.materialized = {}

        # { symbol => tuple type definition }
        self.tupledefs = {}

        # symbol resolution
        self.resolving_symbols = {}

        self.common_subexpression_elim = cse

        self.current_pipeline_properties = {}
        self.current_pipeline_precode = []
        self.current_pipeline_postcode = []

        self.sequence_wait_statements = set()

        self.in_loop = False
        # this is a set for cases where we might generate recycle twice
        # as in operators above a union or other streaming binary operator
        self.loop_recycle_codes = set()
        self.loop_pipeline_codes = []

    def recordCodeWhenInLoop(self, code):
        if self.in_loop:
            self.loop_recycle_codes.add(code)

    def enterLoop(self):
        assert not self.in_loop, "Nested loops not supported"
        self.loop_recycle_codes = set()
        self.loop_pipeline_codes = []
        self.in_loop = True

    def exitLoop(self):
        assert self.in_loop, "Exiting loop but not in one"
        self.in_loop = False
        return self.loop_pipeline_codes, self.loop_recycle_codes

    def setPipelineProperty(self, key, value):
        LOG.debug("set %s in %s" % (key, self.current_pipeline_properties))
        self.current_pipeline_properties[key] = value

    def addToPipelinePropertyList(self, key, value):
        current = self.current_pipeline_properties.get(key, [])
        assert isinstance(current, list), "cannot add to non-list property"
        if len(current) == 0:
            self.current_pipeline_properties[key] = current

        current.append(value)

    def addToPipelinePropertySet(self, key, value):
        current = self.current_pipeline_properties.get(key, set())
        assert isinstance(current, set), "cannot add to non-set property"
        if len(current) == 0:
            self.current_pipeline_properties[key] = current

        current.add(value)

    def getPipelineProperty(self, key):
        LOG.debug("get %s from %s" % (key, self.current_pipeline_properties))
        return self.current_pipeline_properties[key]

    def checkPipelineProperty(self, key):
        """
        Like getPipelineProperty but returns None if no property is found
        """
        LOG.debug("get(to check) %s from %s"
                  % (key, self.current_pipeline_properties))
        return self.current_pipeline_properties.get(key)

    def createUnresolvedSymbol(self):
        name = gensym()
        rs = ResolvingSymbol(name)
        self.resolving_symbols[name] = None
        return rs

    def resolveSymbol(self, rs, value):
        self.resolving_symbols[rs.getName()] = value

    def resolveCounterSymbol(self, rs, increment=1):
        """For late-resolved symbols that are incremented integers whose
        value is unknown until a later compilation step"""

        # Only store strings in resolving_symbols to maintain symbol is
        # a string invariant. We just convert back and forth here to increment
        if self.resolving_symbols[rs.getName()] is None:
            self.resolving_symbols[rs.getName()] = '0'

        self.resolving_symbols[rs.getName()] = str(
            int(self.resolving_symbols[rs.getName()]) + increment)

    def addDeclarations(self, d):
        self.declarations += d

    def addDeclarationsUnresolved(self, d):
        """
        Ordered in the code after the regular declarations
        just so that any name dependences already have been declared
        ALTERNATIVE: split decls into forward decls and definitions
        """
        self.declarations_later += d

    def addInitializers(self, i):
        self.initializers += i

    def addCleanups(self, i):
        self.cleanups += i

    def addSeqWaitStatement(self, c):
        self.sequence_wait_statements.add(c)

    def getAndFlushSeqWaitStatements(self):
        r = self.sequence_wait_statements
        self.sequence_wait_statements = set()
        return r

    def _append_pipeline_code(self, code):
        if self.in_loop:
            self.loop_pipeline_codes.append(code)
        else:
            self.pipelines.append(code)

    def addPipeline(self, p):
        LOG.debug("output pipeline %s", self.current_pipeline_properties)

        all_p = emitlist(self.current_pipeline_precode) \
            + p \
            + emitlist(self.current_pipeline_postcode)

        pipeline_code = \
            self.language.pipeline_wrap(self.pipeline_count, all_p,
                                        self.current_pipeline_properties)

        # force scan pipelines to go first
        if self.current_pipeline_properties.get('type') == 'scan':
            assert not self.in_loop, "scan pipeline not supported in loop"
            self.scan_pipelines.append(pipeline_code)
        else:
            self._append_pipeline_code(pipeline_code)

        self.pipeline_count += 1
        self.current_pipeline_properties = {}
        self.current_pipeline_precode = []
        self.current_pipeline_postcode = []

    def addCode(self, c):
        """
        Just add code here
        """
        self._append_pipeline_code(c)

    def addPreCode(self, c):
        self.current_pipeline_precode.append(c)

    def addPostCode(self, c):
        self.current_pipeline_postcode.append(c)

    def addPipelineFlushCode(self, c):
        self.flush_pipelines.append(c)

    def getInitCode(self):

        # inits is a set.
        # If this ever becomes a bottleneck when declarations are strings,
        # as in cpp, then resort to at least symbol name deduping.
        # TODO: better would be to mark elements of self.initializers as
        # TODO: "do dedup" or "don't dedup"
        s = set()

        def f(x):
            if x in s:
                return False
            else:
                s.add(x)
                return True

        code = emitlist(filter(f, self.initializers))
        return ResolvingSymbol.substitute(code, self.resolving_symbols)

    def getCleanupCode(self):
        # cleanups is a set.
        # If this ever becomes a bottleneck when declarations are strings,
        # as in cpp, then resort to at least symbol name deduping.
        # TODO: better would be to mark elements of self.cleanups as
        # TODO: "do dedup" or "don't dedup"
        s = set()

        def f(x):
            if x in s:
                return False
            else:
                s.add(x)
                return True

        code = emitlist(filter(f, self.cleanups))
        return ResolvingSymbol.substitute(code, self.resolving_symbols)

    def getDeclCode(self):
        # declarations is a set
        # If this ever becomes a bottleneck when declarations are strings,
        # as in cpp, then resort to at least symbol name deduping.
        s = set()

        def f(x):
            if x in s:
                return False
            else:
                s.add(x)
                return True

        # keep in original order
        code = emitlist(filter(f, self.declarations))
        code += emitlist(filter(f, self.declarations_later))
        return ResolvingSymbol.substitute(code, self.resolving_symbols)

    def getExecutionCode(self):
        assert not self.in_loop, "Bad state, in loop at end of compilation"

        # list -> string
        scan_linearized = emitlist(self.scan_pipelines)

        # Make sure we emitted all the wait statements
        if self.sequence_wait_statements:
            waits_linearized = emitlist(
                list(self.getAndFlushSeqWaitStatements()))
        else:
            waits_linearized = ""

        mem_linearized = \
            emitlist(self.pipelines) + waits_linearized

        flush_linearized = emitlist(self.flush_pipelines)
        scan_linearize_wrap = self.language.group_wrap(gensym(),
                                                       scan_linearized,
                                                       {'type': 'scan'})
        mem_linearize_wrap = self.language.group_wrap(gensym(),
                                                      mem_linearized,
                                                      {'type': 'in_memory'})

        linearized = \
            scan_linearize_wrap + mem_linearize_wrap + flush_linearized

        # substitute all lazily resolved symbols
        resolved = ResolvingSymbol.substitute(linearized,
                                              self.resolving_symbols)

        return resolved

    def lookupExpr(self, expr):

        if self.common_subexpression_elim:
            res = self.materialized.get(expr)
            LOG.debug("lookup subexpression %s -> %s", expr, res)
            return res
        else:
            # if CSE is turned off
            # then always return None for expression matches
            return None

    def saveExpr(self, expr, sym):
        LOG.debug("saving subexpression %s -> %s", expr, sym)
        self.materialized[expr] = sym

    def lookupTupleDef(self, sym):
        return self.tupledefs.get(sym)

    def saveTupleDef(self, sym, tupledef):
        self.tupledefs[sym] = tupledef

    def getCurrentPipelineId(self):
        return self.pipeline_count


class Pipelined(object):

    """
    Trait to provide the compilePipeline method
    for calling into pipeline style compilation.

    This is a mixin class that supports cooperative
    multiple inheritance. To use it properly, put
    it _before_ the base class in the inheritance clause
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, *args):
        self._parent = None
        # Ensure this class follows cooperative multiple inheritance
        super(Pipelined, self).__init__(*args)

    def _markAllParents(self, test_mode=False, **kwargs):
        root = self

        def markChildParent(op):
            for c in op.children():
                c._parent = op
                if test_mode:
                    # Test mode turns on assign-once checks of all
                    # Pipeline subclasses instance variables
                    c._freeze()  # _parent is the last allowed attribute
            return []

        [_ for _ in root.postorder_traversal(markChildParent)]

    __isfrozen = False

    def __setattr__(self, key, value):
        """Overriden to allow objects to turn on assigned-once checks

        While we'd prefer subclasses of Pipelined (more specifically Operator)
        to actually be immutable algebraic datatypes, this is an intermediate
        solution. It was introduced to catch bugs that occur from
        reassigning instance variables that are introduced for compilation
        state (see issue https://github.com/uwescience/raco/issues/477).
        """
        if self.__isfrozen and key in self.assigned_attrs:
            if self.__getattribute__(key) == value:
                LOG.warning(
                    '''reassignment of self.{attr} but ignoring because
                    assigned same value {value}'''.format(
                        attr=key,
                        value=value))
                return
            else:
                raise TypeError(
                    """{obj} is a frozen object; self.{attr} = {oldval};
                    tried to assign {newval}""".format(
                        obj=self,
                        attr=key,
                        oldval=self.__getattribute__(key),
                        newval=value))
        # new set created here rather than __init__ because there
        # may be inconsistency in when Pipelined.__init__ is called relative
        # to assignments to instance variables
        if not hasattr(self, 'assigned_attrs'):
            object.__setattr__(self, 'assigned_attrs', set())
        self.assigned_attrs.add(key)
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

    def parent(self):
        return self._parent

    def run_analyses(self, **kwargs):
        """Intended to be called on the root of the tree"""
        self._markAllParents(**kwargs)

    @classmethod
    @abc.abstractmethod
    def language(cls):
        pass

    @classmethod
    @abc.abstractmethod
    def new_tuple_ref(cls, symbol, scheme):
        pass

    @abc.abstractmethod
    def postorder_traversal(self, func):
        pass

    @abc.abstractmethod
    def produce(self, state):
        """Denotation for producing a tuple"""
        return

    @abc.abstractmethod
    def consume(self, inputTuple, fromOp, state):
        """Denotation for consuming a tuple"""
        return

    def compilePipeline(self, compiler='push', **kwargs):
        # run analyses
        self.run_analyses(**kwargs)

        compilerstate = {'push': CompileState,
                         'iterator': IteratorCompileState
                         }[compiler]
        state = compilerstate(self.language())

        state.addCode(
            self.language().comment("Compiled subplan for %s" % self))

        self.produce(state)

        # state.addCode( self.language().log("Evaluating subplan %s" % self) )

        return state


class IteratorCompileState(CompileState):

    def __init__(self, lang, cse=True):
        super(IteratorCompileState, self).__init__(lang, cse)
        self.iterator_operators = []

    def addOperator(self, code):
        self.iterator_operators.append(code)

    def addPipeline(self, p=None):
        # base class addPipeline takes code from the client,
        # but the IteratorCompileState keeps track of the code itself in
        # self.iterator_operators
        if p is None:  # None indicates an iterator pipeline
            p = self.language.iterators_wrap(
                ''.join(
                    self.iterator_operators),
                self.current_pipeline_properties)

        super(IteratorCompileState, self).addPipeline(p)
        self.iterator_operators = []



from collections import Counter, defaultdict
from raco.backends.sql.test_case import SQLTestCase

import raco.scheme as scheme
import raco.types as types


class TestScheme(SQLTestCase):
    """Test that we can convert Raco schemes to SQL Alchemy schemes"""
    def test_simple_scheme(self):
        sch = scheme.Scheme()
        sch.addAttribute('w', types.FLOAT_TYPE)
        sch.addAttribute('x', types.INT_TYPE)
        sch.addAttribute('y', types.LONG_TYPE)
        sch.addAttribute('z', types.STRING_TYPE)

        self.db.add_table('simple', sch)

        sch2 = self.db.get_scheme('simple')
        self.assertEquals(sch, sch2)


class TestQuery(SQLTestCase):
    """Test actually compiling plans to SQL."""
    def test_catalog(self):
        self.assertEquals(len(self.emp_table),
                          self.db.num_tuples(self.emp_key))

    def test_simple_scan(self):
        query = """x = scan({emp});
        store(x, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter(self.emp_table)
        self.execute(query, expected)

    def test_column_select(self):
        query = """x = scan({emp});
        y = [from x emit $2];
        store(y, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((x[2],) for x in self.emp_table)
        self.execute(query, expected)

    def test_rename_column_select(self):
        query = """x = scan({emp});
        y = [from x emit $0 as a, $0 as b];
        store(y, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((x[0], x[0]) for x in self.emp_table)
        self.execute(query, expected)

    def test_bug_380(self):
        query = """emp = scan({emp});
        x = [from emp emit 3 as test];
        store(x, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter([(3,) for _ in self.emp_table])
        self.execute(query, expected)

    def test_count_query(self):
        query = """x = countall(scan({emp}));
        store(x, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter([(len(self.emp_table),)])
        self.execute(query, expected)

    def test_theta_join_query(self):
        query = """x = scan({emp});
        y = scan({emp});
        z = [from x,y where x.salary < y.salary emit x.*];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((a, b, c, d, e)
                           for (a, b, c, d, e) in self.emp_table
                           for (_, _, _, d2, _) in self.emp_table
                           if d < d2)
        self.execute(query, expected)

    def test_projecting_join_query(self):
        query = """x = scan({emp});
        y = scan({emp});
        z = [from x,y where x.mgr_id = y.id
             emit x.id as emp_id, y.salary as mgr_sal];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        expected = Counter((a, d2)
                           for (a, b, c, d, e) in self.emp_table
                           for (a2, _, _, d2, _) in self.emp_table
                           if e == a2)
        self.execute(query, expected)

    def test_join_multi_agg_query(self):
        query = """x = scan({emp});
        y = scan({emp});
        z = [from x,y where x.salary = y.salary
             emit count(*) as cnt, max(x.salary) as max_sal,
                  min(x.salary) as min_sal];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        join = [(1, d)
                for (_, _, _, d, _) in self.emp_table
                for (_, _, _, d2, _) in self.emp_table
                if d == d2]
        size = len(join)
        max_sal = max([d for (_, d) in join])
        min_sal = min([d for (_, d) in join])
        expected = Counter([(size, max_sal, min_sal)])
        self.execute(query, expected)

    def test_complex_agg_query(self):
        query = """x = scan({emp});
        z = [from x emit dept_id, max(salary) as max_salary];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        with self.assertRaises(NotImplementedError):
            d = defaultdict(int)
            for _, dept_id, _, salary, _ in self.emp_table:
                d[dept_id] = max(d[dept_id], salary)
            expected = Counter(d.items())
            self.execute(query, expected)

    def test_case_query(self):
        query = """x = scan({emp});
        z = [from x emit dept_id, case when salary < 20000 then "small"
                                       else "large" end as salary_type];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        def case(salary):
            if salary < 20000:
                return 'small'
            return 'large'
        expected = Counter([(b, case(d)) for _, b, _, d, _ in self.emp_table])
        self.execute(query, expected)

    def test_multi_case_query(self):
        query = """x = scan({emp});
        z = [from x emit dept_id, case when salary < 20000 then "small"
                                       when salary < 50000 then "medium"
                                       else "large" end as salary_type];
        store(z, OUTPUT);""".format(emp=self.emp_key)

        def case(salary):
            if salary < 20000:
                return 'small'
            if salary < 50000:
                return 'medium'
            return 'large'
        expected = Counter([(b, case(d)) for _, b, _, d, _ in self.emp_table])
        self.execute(query, expected)



import requests


class MyriaError(Exception):
    def __init__(self, err=None):
        if isinstance(err, requests.Response):
            msg = 'Error {} ({})'.format(err.status_code, err.reason)
            if err.text:
                msg = '{}: {}'.format(msg, err.text)
            Exception.__init__(self, msg)
        else:
            Exception.__init__(self, err)



# coding=utf-8
""" Tests for built-in functions"""

from collections import Counter
from raco.python import convert
from raco.algebra import Apply
from python_test import PythonTestCase


class TestFunctions(PythonTestCase):
    def _execute(self, query, expected):
        projection = convert(query, [self.schema])
        self.assertIsNotNone(projection)

        expression = Apply([('out', projection)], self.scan)
        plan = self.get_query(expression)
        return self.check_result(plan, expected)

    def test_worker_id(self):
        self._execute("""lambda t: workerid()""",
                      Counter([(0,)] * len(self.emp_table)))

    def test_random(self):
        self._execute("""lambda t: random() <= 1""",
                      Counter([(True,)] * len(self.emp_table)))

    def test_abs(self):
        self._execute("""lambda t: abs(-t[0]) <= 0""",
                      Counter([(False,)] * len(self.emp_table)))

    def test_fabs(self):
        self._execute("""lambda t: fabs(-t[0]) <= 0.25""",
                      Counter([(False,)] * len(self.emp_table)))

    def test_ceil(self):
        self._execute("""lambda t: ceil(t[0] / 2)""",
                      Counter([(1,), (1,), (2,), (2,), (3,), (3,), (4,)]))

    def test_cosine(self):
        self._execute("""lambda t: cos(0)""",
                      Counter([(1,)] * len(self.emp_table)))

    def test_floor(self):
        self._execute("""lambda t: floor(5.5)""",
                      Counter([(5,)] * len(self.emp_table)))

    def test_log(self):
        self._execute("""lambda t: log(1)""",
                      Counter([(0.0,)] * len(self.emp_table)))

    def test_sine(self):
        self._execute("""lambda t: sin(0)""",
                      Counter([(0.0,)] * len(self.emp_table)))

    def test_square_root(self):
        self._execute("""lambda t: sqrt(4)""",
                      Counter([(2,)] * len(self.emp_table)))

    def test_tangent(self):
        self._execute("""lambda t: tan(0)""",
                      Counter([(0,)] * len(self.emp_table)))

    def test_md5_hash(self):
        self._execute("""lambda t: len(str(md5(t[2])))""",
                      Counter([(19,)] * 6 + [(18,)]))

    def test_length(self):
        self._execute("""lambda t: len(t[2])""",
                      Counter([(9,), (9,), (9,), (12,), (14,), (15,), (20,)]))

    def test_power(self):
        self._execute("""lambda t: pow(2, 3)""",
                      Counter([(8,)] * len(self.emp_table)))

    def test_minimum(self):
        self._execute("""lambda t: min(5, 4)""",
                      Counter([(4,)] * len(self.emp_table)))

    def test_minimum_doubles(self):
        self._execute("""lambda t: min(5.5, 4.5)""",
                      Counter([(4.5,)] * len(self.emp_table)))

    def test_maximum(self):
        self._execute("""lambda t: max(10, 20)""",
                      Counter([(20,)] * len(self.emp_table)))

    def test_maximum_doubles(self):
        self._execute("""lambda t: max(10.5, 20.5)""",
                      Counter([(20.5,)] * len(self.emp_table)))



# coding=utf-8

""" Tests for the 'convert' entry-point function """
import ast
import unittest

from raco.expression import NumericLiteral
from raco.python import convert


class TestConvert(unittest.TestCase):
    def test_string(self):
        f = "lambda: 0"
        e = convert(f, None)
        self.assertEqual(e, NumericLiteral(0))

    def test_ast(self):
        t = ast.parse("lambda: 0")
        e = convert(t, None)
        self.assertEqual(e, NumericLiteral(0))

    def test_lambda(self):
        f = lambda: 0
        e = convert(f, None)
        self.assertEqual(e, NumericLiteral(0))

    def test_function(self):
        def f():
            return 0
        e = convert(f, None)
        self.assertEqual(e, NumericLiteral(0))



from raco import RACompiler
from raco.compile import compile

from raco.backends.radish import GrappaSymmetricHashJoin
from raco.backends.radish import GrappaShuffleHashJoin

import raco.viz as viz

import logging
LOG = logging.getLogger(__name__)


def comment(s):
    return "/*\n%s\n*/\n" % str(s)


def hack_plan(alg, plan):
    # plan hacking
    newRule = None
    if plan == "sym":
        alg.set_join_type(GrappaSymmetricHashJoin)
    elif plan == "shuf":
        alg.set_join_type(GrappaShuffleHashJoin)


def emitCode(query, name, algType, plan=None, emit_print=None, dir='.'):
    if emit_print is not None:
        alg = algType(emit_print)
    else:
        alg = algType()

    hack_plan(alg, plan)

    LOG.info("compiling %s: %s", name, query)

    # Create a compiler object
    dlog = RACompiler()

    # parse the query
    dlog.fromDatalog(query)
    # print dlog.parsed
    LOG.info("logical: %s", dlog.logicalplan)

    print dlog.logicalplan
    logical_dot = viz.operator_to_dot(dlog.logicalplan)
    with open("%s.logical.dot" % (name), 'w') as dwf:
        dwf.write(logical_dot)

    dlog.optimize(target=alg)

    LOG.info("physical: %s", dlog.physicalplan)

    print dlog.physicalplan
    physical_dot = viz.operator_to_dot(dlog.physicalplan)
    with open("%s.physical.dot" % (name), 'w') as dwf:
        dwf.write(physical_dot)

    # generate code in the target language
    code = ""
    code += comment("Query " + query)
    code += compile(dlog.physicalplan)

    fname = '{dir}/{name}.cpp'.format(dir=dir, name=name)
    with open(fname, 'w') as f:
        f.write(code)

    # returns name of code file
    return fname





class RepresentationProperties(object):

    def __init__(
            self,
            hash_partitioned=tuple(),
            sorted=None,
            grouped=None,
            broadcasted=False):
        """
        @param hash_partitioned: None or set of AttributeRefs in hash key
        @param sorted: None or list of (AttributeRefs, ASC/DESC) in sort order
        @param grouped: None or list of AttributeRefs to group by

        None means that no knowledge about the interesting property is
        known
        """

        # TODO: make it a set of sets, representing a conjunction of hashes
        # TODO: for example, after a HashJoin($1=$4) we know h($1) && h($4)
        # TODO:     which is not equivalent to h($1, $4). Currently can only
        # TODO:     represent conjunctions of size 1
        self.hash_partitioned = hash_partitioned
        self.broadcasted = broadcasted

        assert not (len(self.hash_partitioned) > 0 and self.broadcasted), \
            "inconsistent state: cannot be partitioned and broadcasted"

        if sorted is not None or grouped is not None:
            raise NotImplementedError("sorted and grouped not yet supported")

    def __str__(self):
        return "{clazz}(hash: {hash_attrs}, broadcasted: {b})".format(
            clazz=self.__class__.__name__,
            hash_attrs=self.hash_partitioned,
            b=self.broadcasted)

    def __repr__(self):
        return "{clazz}({hp!r}, {sort!r}, {grp!r}, {br!r})".format(
            clazz=self.__class__.__name__,
            hp=self.hash_partitioned,
            sort=None,
            grp=None,
            br=self.broadcasted
        )

    def __eq__(self, other):
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        """Override the default hash behavior
        (that returns the id of the object)"""
        return hash(tuple(sorted(self.__dict__.items())))



from abc import ABCMeta, abstractmethod

import logging
from raco.expression.visitor import ExpressionVisitor

LOG = logging.getLogger(__name__)


class Algebra(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def opt_rules(self, **kwargs):
        raise NotImplementedError("{op}.opt_rules()".format(op=type(self)))


class Language(object):
    __metaclass__ = ABCMeta

    EQ = "=="
    NEQ = "!="
    PLUS = "+"
    MINUS = "-"
    DIVIDE = "/"
    TIMES = "*"
    IDIVIDE = "//"
    GT = ">"
    LT = "<"
    GTEQ = ">="
    LTEQ = "<="
    MOD = "%"
    LIKE = "like"

    # By default, reuse scans
    reusescans = True

    @staticmethod
    def preamble(query=None, plan=None):
        return ""

    @staticmethod
    def postamble(query=None, plan=None):
        return ""

    @staticmethod
    def body(compileResult):
        return compileResult

    @classmethod
    def compile_stringliteral(cls, value):
        return '"%s"' % value

    @staticmethod
    def log(txt):
        """Emit code that will generate a log message at runtime. Defaults to
        nothing."""
        return ""

    @classmethod
    def compile_numericliteral(cls, value):
        return '%s' % value

    @classmethod
    def compile_attribute(cls, attr, **kwargs):
        return attr.compile()

    @classmethod
    def conjunction(cls, *args):
        return cls.expression_combine(args, operator="and")

    @classmethod
    def disjunction(cls, *args):
        return cls.expression_combine(args, operator="or")

    @classmethod
    def compile_expression(cls, expr, **kwargs):
        compilevisitor = CompileExpressionVisitor(cls, **kwargs)
        expr.accept(compilevisitor)
        return compilevisitor.getresult()

    @classmethod
    def expression_combine(cls, args, operator="and"):
        """Combine the given arguments using the specified infix operator"""
        delim = " %s " % operator
        return delim.join(args)


class CompileExpressionVisitor(ExpressionVisitor):

    def __init__(self, language, **kwargs):
        self.language = language
        self.combine = language.expression_combine
        self.stack = []
        self.kwargs = kwargs

    def getresult(self):
        assert len(self.stack) == 1, \
            "stack is size {0} != 1".format(len(self.stack))
        return self.stack.pop()

    def __visit_BinaryOperator__(self, binaryexpr):
        right = self.stack.pop()
        left = self.stack.pop()
        return left, right

    def appendbinop(self, binaryexpr, languageop):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        val = self.combine([left, right], operator=languageop)
        self.stack.append(val)

    def visit_NOT(self, unaryexpr):
        inputexpr = self.stack.pop()
        self.stack.append(self.language.negation(inputexpr))

    def visit_AND(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.language.conjunction(left, right))

    def visit_OR(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.language.disjunction(left, right))

    def visit_EQ(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.EQ)

    def visit_NEQ(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.NEQ)

    def visit_GT(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.GT)

    def visit_LT(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.LT)

    def visit_GTEQ(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.GTEQ)

    def visit_LTEQ(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.LTEQ)

    def visit_LIKE(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.LIKE)

    def visit_NamedAttributeRef(self, named):
        self.stack.append(
            self.language.compile_attribute(
                named,
                **self.kwargs))

    def visit_UnnamedAttributeRef(self, unnamed):
        LOG.debug("expr %s is UnnamedAttributeRef", unnamed)
        self.stack.append(
            self.language.compile_attribute(
                unnamed,
                **self.kwargs))

    def visit_NumericLiteral(self, numericliteral):
        self.stack.append(self.language.compile_numericliteral(numericliteral))

    def visit_StringLiteral(self, stringliteral):
        self.stack.append(self.language.compile_stringliteral(stringliteral))

    def visit_DIVIDE(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.DIVIDE)

    def visit_PLUS(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.PLUS)

    def visit_MINUS(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.MINUS)

    def visit_IDIVIDE(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.IDIVIDE)

    def visit_MOD(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.MOD)

    def visit_TIMES(self, binaryexpr):
        self.appendbinop(binaryexpr, self.language.TIMES)

    def visit_NEG(self, unaryexpr):
        inputexpr = self.stack.pop()
        self.stack.append(self.language.negative(inputexpr))

    def visit_Case(self, caseexpr):
        if caseexpr.else_expr is not None:
            else_compiled = self.stack.pop()

        when_compiled = []
        for _ in range(len(caseexpr.when_tuples)):
            thenexpr, ifexpr = self.stack.pop(), self.stack.pop()
            when_compiled.insert(0, (ifexpr, thenexpr))

        self.stack.append(
            self.language.conditional(
                when_compiled,
                else_compiled))

    def visit_NamedStateAttributeRef(self, attr):
        self.stack.append(self.language.compile_attribute(attr, **self.kwargs))

    def visit_UnaryFunction(self, expr):
        inputexpr = self.stack.pop()
        self.stack.append(
            self.language.function_call(
                type(expr).__name__,
                inputexpr))

    def visit_BinaryFunction(self, expr):
        left, right = self.__visit_BinaryOperator__(expr)
        self.stack.append(
            self.language.function_call(
                type(expr).__name__,
                left,
                right))

    def visit_CustomBinaryFunction(self, expr):
        left, right = self.__visit_BinaryOperator__(expr)
        self.stack.append(
            self.language.function_call(
                expr.name,
                left,
                right, custom=True))

    def visit_NaryFunction(self, expr):
        arglist = []
        for _ in range(len(expr.operands)):
            arglist.insert(0, self.stack.pop())
        self.stack.append(
            self.language.function_call(
                type(expr).__name__,
                *arglist))

    def visit_CustomZeroaryFunction(self, expr):
        self.stack.append(
            self.language.function_call(expr.name, custom=True))

    def visit_CAST(self, expr):
        inputexpr = self.stack.pop()
        self.stack.append(self.language.cast(expr._type, inputexpr))



# coding=utf-8
""" Tests for expressions with invalid syntax """

from raco.python import convert
from raco.python.exceptions import PythonConvertException, \
    PythonSyntaxException, PythonArgumentException, \
    PythonUnsupportedOperationException
from python_test import PythonTestCase


class TestSyntax(PythonTestCase):
    def test_invalid_attribute(self):
        query = """lambda t: t.foo == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_invalid_dotted_attribute(self):
        query = """lambda t: t[0].foo == 6"""
        self.assertRaises(PythonUnsupportedOperationException,
                          lambda: convert(query, [self.schema]))

    def test_syntax_error(self):
        query = """lambda t: t[0] = 6"""
        self.assertRaises(PythonSyntaxException,
                          lambda: convert(query, [self.schema]))

    def test_undefined_token(self):
        query = """lambda t: foo[0] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_out_of_range_index(self):
        query = """lambda t: t[999] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_negative_index(self):
        query = """lambda t: t[-1] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_slice(self):
        query = """lambda t: t[1:2] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_substring_negative_index(self):
        query = """lambda t: t.id[-5] == 'x'"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_substring_step(self):
        query = """lambda t: t.id[1:10:3] == 'x'"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_unrecognized_function(self):
        query = """lambda t: foo(t.id) == 1"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_unrecognized_keyword(self):
        query = """lambda t: t.id == foo"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_mismatched_parenthesis(self):
        query = """lambda t: (t[999] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))

    def test_partial_keyword(self):
        query = """foolambda t: t[999] == 6"""
        self.assertRaises(PythonConvertException,
                          lambda: convert(query, [self.schema]))



from abc import ABCMeta, abstractmethod
from collections import Counter


class DatalogPlatformTest(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def check(self, query):
        raise NotImplementedError("{t}.check()".format(t=type(self)))

    @abstractmethod
    def check_file(self, query, name):
        raise NotImplementedError("{t}.check_file()".format(t=type(self)))

    # Run these tests manually with
    #    `python c_test_environment/clang_datalog_tests.py` from `raco/`
    #
    # Currently Grappa tests are disabled unless invoked manually.

    def test_scan(self):
        self.check("A(s1) :- T1(s1)", "scan")

    def test_select(self):
        self.check("A(s1) :- T1(s1), s1>5", "select")

    def test_join(self):
        self.check("A(s1,o2) :- T3(s1,p1,o1), R3(o2,p1,o2)", "join")

    def test_select_conjunction(self):
        self.check("A(s1) :- T1(s1), s1>0, s1<10", "select_conjunction")

    def test_two_var_select(self):
        self.check("A(s1,s2) :- T2(s1,s2), s1<9, s2<9", "two_var_select")

    def test_self_join(self):
        self.check("A(a,b) :- R2(a,b), R2(a,c)", "self_join")

    def test_two_path(self):
        self.check("A(a,b,c) :- R2(a,b), S2(b,c)", "two_path")

    def test_two_hop(self):
        self.check("A(a,c) :- R2(a,b), S2(b,c)", "two_hop")

    def test_three_path(self):
        self.check("A(a,b,c) :- R2(a,b), S2(b,c), T2(c,d)", "three_path")

    def test_self_three_path(self):
        self.check("A(a,b,c) :- R2(a,b), R2(b,c), R2(c,d)", "self_three_path"),

    def test_directed_triangles(self):
        self.check(
            "A(a,b,c) :- R2(a,b), S2(b,c), T2(c,a)",
            "directed_triangles"),

    def test_directed_squares(self):
        self.check(
            "A(a,b,c,d) :- R2(a,b), S2(b,c), T2(c,d), R3(d,a,x)",
            "directed_squares"),

    def test_select_then_join(self):
        self.check(
            "A(s1,s2,s3) :- T3(s1,s2,s3), R2(s3,s4), s1<s2, s4<9",
            "select_then_join"),

    def test_unionall(self):
        self.check("""A(s1) :- T1(s1)
    A(s1) :- R1(s1)""", "unionall")

    def test_swap(self):
        self.check("A(y,x) :- R2(x,y)", "swap"),

    def test_apply(self):
        self.check("""A(x,y) :- T2(x,y)
    B(a) :- A(z,a)""", "apply")

    def test_apply_and_self_join(self):
        self.check("""A(x,z) :- T3(x,y,z), y < 4
    B(x,t) :- A(x,z), A(z,t)""", "apply_and_self_join")

    def test_unionall_apply_and_self_join(self):
        self.check("""A(x,y) :- T2(x,y), R1(x), y < 4
            A(x,y) :- R2(x,y), T1(x)
    B(x,z,t) :- A(x,z), A(z,t)""", "unionall_apply_and_self_join")

    def test_unionall_of_join(self):
        self.check("""A(s1,s2) :- T2(s1,s2)
    A(s1,s2) :- R2(s1,s3), T2(s3,s2)""", "unionall_of_join")

    def test_unionall_then_join(self):
        self.check("""A(s1,s2) :- T2(s1,s2)
    A(s1,s2) :- R2(s1,s2)
    B(s1) :- A(s1,s2), S1(s1)""", "unionall_then_join")

    def test_join_of_two_unionalls(self):
        self.check("""A(s1,s2) :- T2(s1,s2)
    A(s1,s2) :- R2(s1,s2)
    B(s1) :- A(s1,s2), A(s1,s3)""", "join_of_two_unionalls")

    def test_join_swap_indexing(self):
        self.check(
            """A(a,h,y) :- T3(a,b,c), R3(x, y, z), S3(g,h,j), z=c, j=x""",
            "join_swap_indexing")

    def test_head_scalar_op(self):
        self.check("""A(a+b) :- R2(a,b)""", "head_scalar_op")

    def test_aggregate_sum(self):
        self.check("""A(SUM(a)) :- R1(a)""", "aggregate_sum")

    def test_aggregate_count(self):
        self.check("""A(COUNT(a)) :- R1(a)""", "aggregate_count")

    def test_aggregate_count_group_one(self):
        self.check(
            """A(b, COUNT(a)) :- R2(a,b)""",
            "aggregate_count_group_one")

    def test_aggregate_count_group_one_notgroup_one(self):
        self.check(
            """A(b, COUNT(a)) :- R3(a,b,c)""",
            "aggregate_count_group_one_notgroup_one")

    def test_aggregate_count_group_one_notgroup_filtered_one(self):
        self.check(
            """A(b, COUNT(a)) :- R3(a,b,c), c<5""",
            "aggregate_count_group_one_notgroup_filtered_one")

    def test_aggregate_of_binop(self):
        self.check("""A(SUM(a+b)) :- R2(a,b)""", "aggregate_of_binop")

    def test_join_of_aggregate_of_join(self):
        self.check("""A(SUM(a), c) :- R2(a,b), T2(b,c)
                      B(x, y) :- A(x, z), S2(z, y)""",
                   "join_of_aggregate_of_join")

    def test_common_index_allowed(self):
        """introduced for #250"""
        self.check(
            """A(a,b,c,d) :- T2(a,b), R2(a,c), R2(a,d)""",
            "common_index_allowed")

    def test_common_index_disallowed(self):
        """introduced for #250"""
        self.check(
            """A(a,b,c,d) :- T2(a,b), R2(a,c), R2(d,a)""",
            "common_index_disallowed")

    def test_file_store(self):
        self.check_file("""store(a,b) :- R2(a,b)""", "store")

    def test_file_fewer_col_store(self):
        self.check_file("""few_col_store(a) :- R2(a,3)""", "few_col_store")

    def test_file_more_col_store(self):
        self.check_file(
            """more_col_store(a,b,c) :- R2(a,b), S2(b,c), T2(c,a)""",
            "more_col_store")

    def test_file_no_tuple_store(self):
        self.check_file("""zero_store(a) :- R2(a,11)""", "zero_store")

import raco.scheme as scheme
import raco.types as types
import raco.myrial.myrial_test as myrial_test


class MyriaLPlatformTestHarness(myrial_test.MyrialTestCase):
    __metaclass__ = ABCMeta

    def setUp(self):
        super(MyriaLPlatformTestHarness, self).setUp()

        self.tables = {}
        for name in ['R', 'S', 'T', 'I', 'D', 'C']:
            for width in [1, 2, 3]:
                tablename = "%s%d" % (name, width)
                fullname = "public:adhoc:%s" % tablename
                self.tables[tablename] = fullname

                if name == 'D':
                    rest_type = types.DOUBLE_TYPE
                elif name == 'C':
                    rest_type = types.STRING_TYPE
                else:
                    rest_type = types.LONG_TYPE

                one = [("a", types.LONG_TYPE)]
                two = one + [("b", rest_type)]
                three = two + [("c", rest_type)]
                # ingest fake data; data is already generated separately for
                # now
                if width == 1:
                    self.db.ingest(fullname, Counter(), scheme.Scheme(one))
                elif width == 2:
                    self.db.ingest(fullname, Counter(), scheme.Scheme(two))
                else:
                    self.db.ingest(fullname, Counter(), scheme.Scheme(three))

    @abstractmethod
    def check(self, query, name):
        pass


class MyriaLPlatformTests(object):

    def myrial_from_sql(self, tables, name):
        code = [
            "{t} = scan({tab});".format(
                t=t,
                tab=self.tables[t]) for t in tables]

        with open("c_test_environment/testqueries/%s.sql" % name) as f:
            code.append("out = %s" % f.read())

        code.append("store(out, OUTPUT);")

        return '\n'.join(code)

    def check_sub_tables(self, query, name, **kwargs):
        self.check(query % self.tables, name, **kwargs)

    def test_scan(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        STORE(T1, OUTPUT);
        """, "scan")

    def test_sink(self):
        """
        Sink still prints on verbose=2, so same as store for testing method
        """
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        SINK(T1);
        """, "scan")

    def test_select(self):
        q = self.myrial_from_sql(['T1'], "select")
        self.check(q, "select")

    def test_join(self):
        self.check_sub_tables("""
        T3 = SCAN(%(T3)s);
        R3 = SCAN(%(R3)s);
        out = JOIN(T3, b, R3, b);
        out2 = [FROM out WHERE $3 = $5 EMIT $0, $3];
        STORE(out2, OUTPUT);
        """, "join")

    def test_select_conjunction(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        out = [FROM T1 WHERE a>0 and a<10 EMIT a];
        STORE(out, OUTPUT);
        """, "select_conjunction")

    def test_two_var_select(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        out = [FROM T2 WHERE a<9 and b<9 EMIT a,b];
        STORE(out, OUTPUT);
        """, "two_var_select")

    def test_self_join(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        j = JOIN(R2, a, R2, a);
        out = [FROM j EMIT $0, $1];
        STORE(out, OUTPUT);
        """, "self_join")

    def test_two_path(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        j = JOIN(R2, b, S2, a);
        out = [FROM j EMIT $0, $1, $3];
        STORE(out, OUTPUT);
        """, "two_path")

    def test_two_hop(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        j = JOIN(R2, b, S2, a);
        out = [FROM j EMIT $0, $3];
        STORE(out, OUTPUT);
        """, "two_hop")

    def test_three_path(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        T2 = SCAN(%(T2)s);
        j1 = JOIN(R2, b, S2, a);
        j2 = JOIN(j1, $3, T2, a);
        out = [FROM j2 EMIT $0, $1, $3];
        STORE(out, OUTPUT);
        """, "three_path")

    def test_self_three_path(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        j1 = JOIN(R2, b, R2, a);
        j2 = JOIN(j1, $3, R2, a);
        out = [FROM j2 EMIT $0, $1, $3];
        STORE(out, OUTPUT);
        """, "self_three_path")

    def test_directed_triangles(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        T2 = SCAN(%(T2)s);
        j1 = JOIN(R2, b, S2, a);
        j2 = JOIN(j1, $3, T2, a);
        out = [FROM j2 WHERE $0 = $5 EMIT $0, $1, $3];
        STORE(out, OUTPUT);
        """, "directed_triangles")

    def test_directed_squares(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        T2 = SCAN(%(T2)s);
        R3 = SCAN(%(R3)s);
        out = select R2.$0 as a, S2.$0 as b, T2.$0 as c, R3.$0 as d
              from R2, S2, T2, R3
              where R2.$1 = S2.$0 and S2.$1 = T2.$0 and T2.$1 = R3.$0
                and R3.$1 = R2.$0;
        STORE(out, OUTPUT);
        """, "directed_squares")

    def test_select_then_join(self):
        self.check_sub_tables("""
        T3 = SCAN(%(T3)s);
        R2 = SCAN(%(R2)s);
        j = JOIN(T3, c, R2, a);
        out = [FROM j WHERE $0 < $1 and $4 < 9 EMIT $0,$1,$2];
        STORE(out, OUTPUT);
        """, "select_then_join")

    def test_unionall(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        R1 = SCAN(%(R1)s);
        un = UNIONALL(T1, R1);
        STORE(un, OUTPUT);
        """, "unionall")

    def test_unionall_3(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        R1 = SCAN(%(R1)s);
        S1 = SCAN(%(S1)s);
        un = UNIONALL(T1, R1, S1);
        STORE(un, OUTPUT);
        """, "unionall_3")

    def test_swap(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        out = [FROM R2 EMIT b, a];
        STORE(out, OUTPUT);
        """, "swap")

    def test_apply(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        interm = [FROM T2 EMIT $0, $1];
        out = [FROM interm EMIT $1];
        STORE(out, OUTPUT);
        """, "apply")

    def test_idivide(self):
        q = self.myrial_from_sql(['T2'], "idivide").replace("/", "//")
        self.check(q, "idivide")

    def test_apply_and_self_join(self):
        q = self.myrial_from_sql(['T3'], "apply_and_self_join")
        self.check(q, "apply_and_self_join")

    def test_unionall_apply_and_self_join(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        R1 = SCAN(%(R1)s);
        R2 = SCAN(%(R2)s);
        T1 = SCAN(%(T1)s);
        Aj1 = JOIN(T2, a, R1, a);
        A1 = [FROM Aj1 WHERE $1 < 4 EMIT $0 as x, $1 as y];
        Aj2 = JOIN(R2, a, T1, a);
        A2 = [FROM Aj2 EMIT $0 as x, $1 as y];
        AU = UNIONALL(A1, A2);
        B = JOIN(AU, $1, AU, $0);
        out = [FROM B EMIT $0 as x, $1 as y, $3 as t];
        STORE(out, OUTPUT);""", "unionall_apply_and_self_join")

    def test_unionall_of_join(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        R2 = SCAN(%(R2)s);
        A1 = [FROM T2 EMIT a, b];
        A2 = JOIN(R2, b, T2, a);
        A2b = [FROM A2 EMIT $0, $3];
        out = UNIONALL(A1, A2b);
        STORE(out, OUTPUT);""",
                              "unionall_of_join")

    def test_unionall_then_join(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        R2 = SCAN(%(R2)s);
        S1 = SCAN(%(S1)s);
        A = UNIONALL(T2, R2);
        B = JOIN(A, $0, S1, $0);
        out = [FROM B EMIT $0];
        STORE(out, OUTPUT);
        """, "unionall_then_join")

    def test_join_of_two_unionalls(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        R2 = SCAN(%(R2)s);
        A = UNIONALL(T2, R2);
        B = JOIN(A, $0, A, $0);
        out = [FROM B EMIT $0];
        STORE(out, OUTPUT);
        """, "join_of_two_unionalls")

    def test_join_swap_indexing(self):
        q = self.myrial_from_sql(["T3", "S3", "R3"], "join_swap_indexing")
        self.check(q, "join_swap_indexing")

    def test_head_scalar_op(self):
        q = self.myrial_from_sql(["R2"], "head_scalar_op")
        self.check(q, "head_scalar_op")

    def test_aggregate_sum(self):
        q = self.myrial_from_sql(["R1"], "aggregate_sum")
        self.check(q, "aggregate_sum")

    def test_aggregate_count(self):
        q = self.myrial_from_sql(["R1"], "aggregate_count")
        self.check(q, "aggregate_count")

    def test_aggregate_min(self):
        q = self.myrial_from_sql(["T2"], "aggregate_min")
        self.check(q, "aggregate_min")

    def test_aggregate_max(self):
        q = self.myrial_from_sql(["T2"], "aggregate_max")
        self.check(q, "aggregate_max")

    def test_aggregate_count_group_one(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        out = [FROM R2 EMIT b, COUNT(a)];
        STORE(out, OUTPUT);
        """, "aggregate_count_group_one")

    def test_aggregate_count_group_one_notgroup_one(self):
        self.check_sub_tables("""
        R3 = SCAN(%(R3)s);
        out = [FROM R3 EMIT b, COUNT(a)];
        STORE(out, OUTPUT);
        """, "aggregate_count_group_one_notgroup_one")

    def test_aggregate_count_group_one_notgroup_filtered_one(self):
        self.check_sub_tables("""
        R3 = SCAN(%(R3)s);
        out = [FROM R3 WHERE R3.c < 5 EMIT b, COUNT(a)];
        STORE(out, OUTPUT);
        """, "aggregate_count_group_one_notgroup_filtered_one")

    def test_aggregate_of_binop(self):
        q = self.myrial_from_sql(["R2"], "aggregate_of_binop")
        self.check(q, "aggregate_of_binop")

    def test_join_of_aggregate_of_join(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        Ap = JOIN(R2, b, T2, a);
        A = [FROM Ap EMIT SUM($0), $3];
        B = JOIN(A, $1, S2, a);
        out = [FROM B EMIT $0, $3];
        STORE(out, OUTPUT);
        """, "join_of_aggregate_of_join")

    def test_join_two_types(self):
        """
        Makes sure that key types are not mixed up.
        A related bug came up when input attribute types were
        inadvertantly computed relative to a hashjoin schema instead of
        the schema of its inputs
        """
        q = self.myrial_from_sql(["C3", "R3"], "join_two_types")
        self.check(q, "join_two_types")

    def test_join_of_two_aggregates(self):
        """
        Goal is to force aggregate result to be insert side of a hash join.
        """
        self.check_sub_tables("""
        D2 = SCAN(%(D2)s);
        D3 = SCAN(%(D3)s);
        agg1 = select a, MIN(b) as mb from D2;
        agg2 = select a, MIN(b) as mb from D3;
        out = select agg1.a, agg2.a
            from agg1, agg2
            where agg1.mb = agg2.mb;
        STORE(out, OUTPUT);
        """, "test_join_of_two_aggregates")

    def test_common_index_allowed(self):
        q = self.myrial_from_sql(["R2", "T2"], "common_index_allowed")
        self.check(q, "common_index_allowed")

    def test_common_index_disallowed(self):
        q = self.myrial_from_sql(["R2", "T2"], "common_index_disallowed")
        self.check(q, "common_index_disallowed")

    def test_matrix_mult(self):
        self.check_sub_tables("""
        T = scan(%(T2)s);
        T1 = T;
        T2 = T;
        MM = [from T1, T2
              where T1.$1 = T2.$0
              emit T1.$0 as src, T2.$1 as dst, count(T1.$0)];
        STORE(MM, OUTPUT);
        """, "matrix_mult")

    def test_two_join_switch(self):
        self.check_sub_tables("""
        R3 = SCAN(%(R3)s);
        S3 = SCAN(%(S3)s);
        T3 = SCAN(%(T3)s);
        J1 = JOIN(R3, $2, S3, $1);
        J2 = JOIN(J1, $3, T3, $0);
        P = [FROM J2 WHERE $0>1 and $3>2 and $6>3 EMIT $0, $8];
        STORE(P, OUTPUT);
        """, "two_join_switch", SwapJoinSides=True)

    def test_select_double(self):
        q = self.myrial_from_sql(["D3"], "select_double")
        self.check(q, "select_double")

    def test_project_string(self):
        q = self.myrial_from_sql(["C3"], "project_string")
        self.check(q, "project_string")

    def test_select_string(self):
        q = self.myrial_from_sql(["C3"], "select_string")
        self.check(q, "select_string")

    def test_join_string_val(self):
        q = self.myrial_from_sql(["C2", "T2"], "join_string_val")
        self.check(q, "join_string_val")

    def test_join_string_key(self):
        q = self.myrial_from_sql(["C3", "C3"], "join_string_key")
        self.check(q, "join_string_key")

    def test_groupby_string_key(self):
        self.check_sub_tables("""
        C2 = SCAN(%(C2)s);
        P = [FROM C2 EMIT SUM($0), $1];
        STORE(P, OUTPUT);
        """, "groupby_string_key")

    def test_groupby_string_multi_key(self):
        self.check_sub_tables("""
        C3 = SCAN(%(C3)s);
        P = [FROM C3 EMIT SUM($0), $1, $2];
        STORE(P, OUTPUT);
        """, "groupby_string_multi_key")

    def test_select_string_literal(self):
        q = self.myrial_from_sql(["C3"], "select_string_literal")
        self.check(q, "select_string_literal")

    def test_aggregate_string(self):
        self.check_sub_tables("""
       C3 = SCAN(%(C3)s);
       P = [FROM C3 EMIT $0, COUNT($1)];
       STORE(P, OUTPUT);
       """, "aggregate_string")

    def test_countstar_string(self):
        self.check_sub_tables("""
       C3 = SCAN(%(C3)s);
       P = [FROM C3 EMIT COUNT($1)];
       STORE(P, OUTPUT);
       """, "countstar_string")

    def test_like_begin_end(self):
        q = self.myrial_from_sql(["C2"], "like_begin_end")
        self.check(q, "like_begin_end")

    def test_like_middle(self):
        q = self.myrial_from_sql(["C2"], "like_middle")
        self.check(q, "like_middle")

    def test_like_end(self):
        q = self.myrial_from_sql(["C2"], "like_end")
        self.check(q, "like_end")

    def test_like_begin(self):
        q = self.myrial_from_sql(["C2"], "like_begin")
        self.check(q, "like_begin")

    def test_aggregate_double(self):
        self.check_sub_tables("""
        D2 = SCAN(%(D2)s);
        out = select a, SUM(b) from D2;
        STORE(out, OUTPUT);
        """, "aggregate_double")

    def test_aggregate_of_binop_double(self):
        self.check_sub_tables("""
        D3 = SCAN(%(D3)s);
        out = select a, MAX(b-c) from D3;
        STORE(out, OUTPUT);
        """, "aggregate_of_binop_double")

    def test_aggregate_of_binop_no_key_unionall_double(self):
        self.check_sub_tables("""
        D3 = SCAN(%(D3)s);
        ma = select MAX(b-c) from D3;
        mi = select MIN(c-b) from D3;
        out = UNIONALL(ma, mi);
        STORE(out, OUTPUT);
        """, "aggregate_of_binop_no_key_unionall_double")

    def test_turn_off_optimizations(self):
        # just testing that it doesn't break anything
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        out = [FROM T1 WHERE a>0 and a<10 EMIT a];
        STORE(out, OUTPUT);
        """, "select_conjunction",
                              no_SplitSelects=True,
                              no_MergeSelects=True,
                              no_PushSelects=True)

    def test_iterator_select(self):
        q = self.myrial_from_sql(['T1'], "select")
        self.check(q, "select", compiler='iterator')

    def test_iterator_apply(self):
        self.check_sub_tables("""
        T2 = SCAN(%(T2)s);
        interm = [FROM T2 EMIT $0, $1];
        out = [FROM interm EMIT $1];
        STORE(out, OUTPUT);
        """, "apply", compiler='iterator')

    def test_iterator_join(self):
        self.check_sub_tables("""
        T3 = SCAN(%(T3)s);
        R3 = SCAN(%(R3)s);
        out = JOIN(T3, b, R3, b);
        out2 = [FROM out WHERE $3 = $5 EMIT $0, $3];
        STORE(out2, OUTPUT);
        """, "join", compiler='iterator')

    def test_iterator_aggregate_sum(self):
        q = self.myrial_from_sql(["R1"], "aggregate_sum")
        self.check(q, "aggregate_sum", compiler='iterator')

    def test_iterator_aggregate_count_group_one(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        out = [FROM R2 EMIT b, COUNT(a)];
        STORE(out, OUTPUT);
        """, "aggregate_count_group_one", compiler='iterator')

    def test_symmetric_hash_join(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        T2 = SCAN(%(T2)s);
        j1 = JOIN(R2, b, S2, a);
        j2 = JOIN(j1, $3, T2, a);
        out = [FROM j2 WHERE $0 = $5 EMIT $0, $1, $3];
        STORE(out, OUTPUT);
        """, "directed_triangles", join_type='symmetric_hash')

    def test_symmetric_hash_join_then_aggregate(self):
        """to fix https://github.com/uwescience/raco/issues/477, caused by
        multiple instances of a pipeline"""
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        T2 = SCAN(%(T2)s);
        j2 = select * from R2, S2, T2 where R2.b=S2.a and S2.a=T2.a;
        a = select SUM($0), $1 from j2;
        STORE(a, OUTPUT);
        """, "join_then_aggregate", join_type='symmetric_hash')

    def test_unionall_then_aggregate(self):
        self.check_sub_tables("""
        R2 = SCAN(%(R2)s);
        S2 = SCAN(%(S2)s);
        u = UNIONALL(R2, S2);
        a = select SUM($0), $1 from u;
        STORE(a, OUTPUT);
        """, "unionall_then_aggregate")

    def test_store_file(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        STORE(T1, OUTPUT);
        """, "scan", emit_print='file')

    def test_q2(self):
        """
        A test resembling sp2bench Q2
        """

        self.check_sub_tables("""
            S = SCAN(R3);
            P = [FROM S T1,
            S T2,
            S T3,
            S T4,
            S T5,
            S T6,
            S T7,
            S T8,
            S T9
WHERE T1.a=T2.a
and T2.a=T3.a
and T3.a=T4.a
and T4.a=T5.a
and T5.a=T6.a
and T6.a=T7.a
and T7.a=T8.a
and T8.a=T9.a
and T1.b = 1 and T1.c > 5
and T2.b = 1
and T3.b = 1
and T4.b = 1
and T5.b = 1
and T6.b = 1
and T7.b = 1
and T8.b = 1
and T9.b = 1
EMIT
T1.a as inproc,
T2.c as author,
T3.c as booktitle,
T4.c as title,
T5.c as proc,
T6.c as ee,
T7.c as page,
T8.c as url,
T9.c as yr];
STORE(P,OUTPUT);""", "q2")
